# -- Future Imports -- (Use with caution, may not work as expected in all cases)
from __future__ import annotations

# -- STL Imports --
from dataclasses import dataclass
import math
import re
from typing import Any, Optional, TypeVar

# -- Library Imports --
from rusttypes.option import Nil, Option, Some

# -- Package Imports --
from parasite.errors import ValidationError
from parasite.type import ParasiteType, _NotFound

K = TypeVar("K")
"""Template type for the key in a dictionary."""


@dataclass
class Boolean(ParasiteType[bool]):
    """
    Parasite type for representing boolean values.

    Inheritance:
        ParasiteType[bool]

    Args:
        _f_optional (bool): Whether the value is optional. Default: False
        _f_nullable (bool): Whether the value can be None. Default: False
        _m_leaniant (tuple[AnyStr, AnyStr]): The regular expressions for the true and false value.
        Default: (r"^(true|1|yes|y)$", r"^(false|0|no|n)$")
        _m_literal (bool | None): The literal value of the boolean. Default: None
    """
    _f_optional: bool = False
    _f_nullable: bool = False
    _f_leaniant: bool = False

    _m_leaniant: tuple[str, str] = (r"^(true|1|yes|y)$", r"^(false|0|no|n)$")
    _m_literal: bool | None = None

    def optional(self) -> Boolean:
        """
        Set the value to be optional.

        Returns:
            Boolean: The updated instance of the class.
        """
        self._f_optional = True
        return self

    def required(self) -> Boolean:
        """
        Set the value to be required.

        Returns:
            Boolean: The updated instance of the class.
        """
        self._f_optional = False
        return self

    def nullable(self) -> Boolean:
        """
        Set the value to be nullable.

        Returns:
            Boolean: The updated instance of the class.
        """
        self._f_nullable = True
        return self

    def non_nullable(self) -> Boolean:
        """
        Set the value to be not nullable.

        Returns:
            Boolean: The updated instance of the class.
        """
        self._f_nullable = False
        return self

    def literal(self, value: bool) -> Boolean:
        """
        Set the literal value of the boolean.

        Args:
            value (bool): The literal value of the boolean.

        Returns:
            Boolean: The updated instance of the class.
        """
        self._m_literal = value
        return self

    def leaniant(self, re_true: Optional[str] = None, re_false: Optional[str] = None) -> Boolean:
        """
        Set the value to be leaniant.

        Args:
            re_true (Optional[str]): The regular expression for the true value. Default: None
            re_false (Optional[str]): The regular expression for the false value. Default: None

        Returns:
            Boolean: The updated instance of the class.
        """
        self._f_leaniant = True

        if re_true is not None:
            self._m_leaniant = (re_true, self._m_leaniant[1])

        if re_false is not None:
            self._m_leaniant = (self._m_leaniant[0], re_false)

        return self

    def parse(self, obj: Any) -> bool:
        # if obj is already a boolean, return it
        if isinstance(obj, bool):
            pass

        # if obj is a string and leaniant mode is active, try to convert it to a boolean
        elif isinstance(obj, str) and self._f_leaniant:
            # if obj is a string, try to convert it to a boolean
            if re.match(self._m_leaniant[0], obj.lower()):
                obj = True

            elif re.match(self._m_leaniant[1], obj.lower()):
                obj = False

            else:
                # raise an error if the value could not be matched to any regex
                raise ValidationError(
                    f"object has to be regex (true: {self._m_leaniant[0]!r}, false: "
                    f"{self._m_leaniant[1]!r}) accepted boolean value, but is '{obj!r}'"
                )

        # if obj is a number, try to convert it to a boolean
        elif isinstance(obj, (int, float)) and self._f_leaniant:
            # first check if the number is an actual number
            if math.isnan(obj):
                obj = False

            elif int(obj) == 1:
                obj = True

            elif int(obj) == 0:
                obj = False

            else:
                raise ValidationError(f"object has to be 1 or 0, but is '{obj!r}'")

        else:
            # raise an error if the value could not be parsed
            raise ValidationError(f"object has to be a boolean, but is '{obj!r}'")

        # if literal is set, check if obj is the literal value
        if self._m_literal is not None and obj != self._m_literal:
            raise ValidationError(f"object has to be {self._m_literal}, but is {obj}")

        return obj

    def find_and_parse(self, parent: dict[K, Any], key: K) -> Option[bool | None]:
        if (value := parent.get(key, _NotFound)) is not _NotFound:
            # if key is found, just package `parse(..)` it into a Some
            if value is not None:
                return Some(self.parse(value))

            # if value is None, check if the value is nullable
            if self._f_nullable:
                return Some(None)

            raise ValidationError(f"key '{key}' cannot be None")

        # if key is not found, return Nil if optional, else raise an error
        if self._f_optional:
            return Nil

        raise ValidationError(f"key '{key}' not found, but is required")



# -- Future Imports -- (Use with caution, may not work as expected in all cases)
from __future__ import annotations

# -- STL Imports --
from dataclasses import dataclass
from typing import Any, TypeVar

# -- Library Imports --
from rusttypes.option import Nil, Option, Some

# -- Package Imports --
from parasite.errors import ValidationError
from parasite.type import ParasiteType, _NotFound

K = TypeVar("K")
"""Template type for the key in a dictionary."""


@dataclass
class Null(ParasiteType[None]):
    """
    Parasite type for representing None values.

    Inheritance:
        ParasiteType[None]

    Args:
        _f_optional (bool): Whether the value is optional. Default: False
    """
    _f_optional: bool = False

    def optional(self) -> Null:
        """
        Makes the value optional, when parsing with `find_and_parse(..)`. Has no effect on
        `parse(..)`. Inverse of `required(..)`.

        Returns:
            Null: modified instance
        """
        self._f_optional = True
        return self

    def required(self) -> Null:
        """
        Makes the value required, when parsing with `find_and_parse(..)`. Has no effect on
        `parse(..)`. Default behavior. Inverse of `optional(..)`.

        Returns:
            Null: modified instance
        """
        self._f_optional = False
        return self

    def parse(self, obj: Any) -> None:
        # if obj is None, return None
        if obj is None:
            return None

        # else raise an error
        raise ValidationError(f"object has to be None, but is '{obj!r}'")

    def find_and_parse(self, parent: dict[K, Any], key: K) -> Option[None]:
        # if key is found, just package `parse(..)` it into a Some
        if (value := parent.get(key, _NotFound)) is not _NotFound:
            return Some(self.parse(value))

        # if key is not found, return Nil if optional, else raise an error
        if self._f_optional:
            return Nil

        raise ValidationError(f"key '{key}' not found, but is required")



# -- Future Imports -- (Use with caution, may not work as expected in all cases)
from __future__ import annotations

# -- STL Imports --
from dataclasses import dataclass, field, replace
from typing import Any, TypeVar

# -- Library Imports --
from rusttypes.option import Nil, Option, Some

# -- Package Imports --
from parasite.errors import ValidationError
from parasite.type import ParasiteType, _NotFound
from parasite.variant import Variant

K = TypeVar("K")
"""Template type for the key in a dictionary."""


@dataclass
class Object(ParasiteType[dict[Any, Any]]):
    """
    Parasite type for representing dictionary values.

    Inheritance:
        ParasiteType[dict]

    Args:
        _f_optional (bool): Whether the value is optional. Default: False
        _f_nullable (bool): Whether the value can be None. Default: False
        _f_strict (bool): Whether the dictionary should be parsed and not allow any other keys to
        exist. Default: False
        _f_strip (bool): Whether the dictionary should be stripped of all keys that are not in the
        _m_items (dict[K, ParasiteType]): The items of the dictionary. Default: {}
    """
    # NOTE: do not move this attribute, this has to be first in the class, as it will break, reading
    # items from the dictionary functionality
    _m_items: dict[Any, ParasiteType] = field(default_factory = lambda: {})

    _f_optional: bool = False
    _f_nullable: bool = False
    _f_strict: bool = False
    _f_strip: bool = False

    def optional(self) -> Object:
        """
        Set the value to be optional.

        Returns:
            Object: The updated instance of the class.
        """
        self._f_optional = True
        return self

    def required(self) -> Object:
        """
        Set the value to be required.

        Returns:
            Object: The updated instance of the class.
        """
        self._f_optional = False
        return self

    def nullable(self) -> Object:
        """
        Set the value to be nullable.

        Returns:
            Object: The updated instance of the class.
        """
        self._f_nullable = True
        return self

    def non_nullable(self) -> Object:
        """
        Set the value to be non-nullable.

        Returns:
            Object: The updated instance of the class.
        """
        self._f_nullable = False
        return self

    def strict(self) -> Object:
        """
        Set the dictionary to be strict.

        Returns:
            Object: The updated instance of the class.
        """
        self._f_strict = True
        return self

    def strip(self) -> Object:
        """
        Set the dictionary to be stripped.

        Returns:
            Object: The updated instance of the class.
        """
        self._f_strip = True
        return self

    def extend(self, other: Object) -> Object:
        """
        Extend the dictionary with another dictionary.

        Args:
            other (Object): The dictionary to extend with.

        Returns:
            Object: The updated instance of the class.
        """
        if not isinstance(other, Object):
            raise ValidationError(f"object has to be a dictionary, but is '{other!r}'")

        self._m_items.update(other._m_items)
        return self

    def merge(self, other: Object) -> Object:
        """
        Merge the dictionary with another dictionary.

        Args:
            other (Object): The dictionary to merge with.

        Returns:
            Object: The updated instance of the class.
        """
        for key, value in other._m_items.items():

            # If the key is in the dictionary, merge the values.
            if key in self._m_items:
                # if both src and dest are objects, merge them
                if isinstance(value, Object) and isinstance(self._m_items[key], Object):
                    self._m_items[key].merge(value)

                # if both src and dest are arrays, merge them
                elif isinstance(value, Variant) and isinstance(self._m_items[key], Variant):
                    for variant in value._m_variants:
                        self._m_items[key].add_variant(variant)

                # if dest is already a variant, add the value to it
                elif isinstance(self._m_items[key], Variant):
                    self._m_items[key].add_variant(value)

                # else, create a new variant and add both values
                else:
                    org = self._m_items[key]
                    self._m_items[key] = Variant().add_variant(org).add_variant(value)

            # If the key is not in the dictionary, add it.
            else:
                self._m_items[key] = value

        return self

    def pick(self, keys: list) -> Object:
        """
        Pick only the keys from the dictionary.

        Args:
            keys (list[K]): The keys to pick.

        Returns:
            Object: New instance of the class with only the picked keys.
        """
        new_obj = replace(self)
        new_obj._m_items = {key: self._m_items[key] for key in keys}
        return new_obj

    def pick_safe(self, keys: list) -> Object:
        """
        Pick only the keys from the dictionary.

        Args:
            keys (list[K]): The keys to pick.

        Returns:
            Object: The updated instance of the class.
        """
        new_obj = replace(self)
        new_obj._m_items = {key: self._m_items[key] for key in keys if key in self._m_items}
        return new_obj

    def omit(self, keys: list) -> Object:
        """
        Omit the keys from the dictionary.

        Args:
            keys (list[K]): The keys to omit.

        Returns:
            Object: New instance of the class with the omitted keys.
        """
        new_obj = replace(self)
        new_obj._m_items = {key: value for key, value in self._m_items.items() if key not in keys}
        return new_obj

    def add_item(self, key: Any, item: ParasiteType) -> Object:
        """
        Add an item to the dictionary.

        Args:
            key (K): The key of the item.
            item (ParasiteType): The item to add.

        Returns:
            Object: The updated instance of the class.
        """
        self._m_items[key] = item
        return self

    def parse(self, obj: Any) -> dict[Any, Any]:
        if not isinstance(obj, dict):
            raise ValidationError(f"object has to be a dictionary, but is '{obj!r}'")

        # If the dictionary should be strict, check if all keys are allowed.
        if self._f_strict:
            for key in obj.keys():
                if key not in self._m_items:
                    raise ValidationError(f"object has the '{key}', but is not allowed to")

        # If the dictionary should be stripped, strip it.
        if self._f_strip:
            obj = {key: value for key, value in obj.items() if key in self._m_items.keys()}

        # Parse the dictionary.
        for key, item in self._m_items.items():
            item.find_and_parse(obj, key).map(lambda x: obj.update({key: x}))

        return obj

    def find_and_parse(self, parent: dict[K, Any], key: K) -> Option[dict[K, Any] | None]:
        if (value := parent.get(key, _NotFound)) is not _NotFound:
            # If key is found, just package `parse(..)` it into a Some.
            if value is not None:
                return Some(self.parse(value))

            # If value is None, check if the value is nullable.
            if self._f_nullable:
                return Some(None)

        # If the value is optional, return a Nil.
        if self._f_optional:
            return Nil

        # If the value is required, raise an error.
        raise ValidationError(f"object has to be a dictionary, but is '{value!r}'")
